from logging import getLogger
from typing import Union, Tuple, Dict, Any
from types import MethodType
from functools import wraps
import json

from click import secho
from flask import request, Flask

from flask_shortcut.util import diff, get_request_data

logger = getLogger(__name__)

#: Shorthand for the return type of something pretending to be a route function.
RESPONSE_ARGS = Tuple[Any, int]


class Shortcut:
    """Object that handles the shortcut rerouting.

    Calling an instance of this class on a function gives that function the
    option to behave differently in non-production environments. The way in
    which the behavior may differ is constrained in two possible ways.

    In the first one, only the arguments for a response are passed to the
    shortcut definition, meaning that the original function is effectively
    disabled and the route will instead just return the shortcut's arguments
    as its only response.

    In the second one, any number of shortcut response arguments are mapped
    to condition-keys. The condition is a string that is used to assert a
    substructure in requests that reach that route, and will only apply its
    respective shortcut-response iff that substructure can be matched.

    If none of the condition can be satisfied, the route will run its
    original logic.

    There are two different ways to register shortcuts, one using decorators
    on the target functions before the are decorated as routes, and the
    other by running the a single wire call after all routes were added.

    Example:

        - Basic app setup:

        >>> app = Flask(__name__)
        >>> short = Shortcut(app)

        - With shortcut decorator:

        >>> @app.route('/my_route', methods=['GET'])
        ... @short.cut(('short_ok', 200))
        ... def my_func():
        ...     return 'ok', 200

        - Equivalent post route-definition wiring:

        >>> @app.route('/my_route', methods=['GET'])
        ... def my_func():
        ...     return 'ok', 200
        >>> short.wire({'/my_route': ('short_ok', 200)})
    """

    def __init__(self, app: Flask):
        self.app = app
        self.exclude = ["production"]
        if app.config.get("SHORTCUT_EXCLUSIONS"):
            exclusions = [x.strip() for x in app.config["SHORTCUT_EXCLUSIONS"].split(",") if x]
            self.exclude.extend(exclusions)
        if app.env not in self.exclude:
            logger.info(f"Setting up flask shortcuts in environment '{app.env}'.")
            secho("   Route shortcuts enabled. Do not do this in any kind of production environment.", fg="red")
        else:
            # make .cut(...) return a wrapper that does nothing
            self.cut = MethodType(lambda _self, mapping: lambda f: f, self)  # type: ignore

    def cut(self, mapping: Union[RESPONSE_ARGS, Dict[str, RESPONSE_ARGS]]):
        """Returns route wrappers.

        Depending on the input argument, a different wrapper will be returned.
        This function can only run in applications that are not listed in the
        `_EXCLUDE` list.

        Args:
            mapping: The mapping that decides which types of shortcuts can be
                offered under which type of condition.
        """
        # wrapper in case only the response is given, assumes that the 'condition' is always True
        def simple_map(f):
            f_name = f"{f.__module__}.{f.__name__}"
            logger.info(f"Adding simple_map shortcut for routing function '{f_name}'.")

            @wraps(f)
            def decorated(*_, **__):
                assert isinstance(mapping, tuple), "Messed up shortcut wiring, abort."  # nosec
                logger.debug(f"Running shortcut for '{f_name}'.")

                response, status = mapping
                return self.app.make_response(response), status

            return decorated

        # wrapper for dict-based mappings
        def dict_map(f):
            f_name = f"{f.__module__}.{f.__name__}"
            logger.info(f"Adding dict_map shortcut for routing function '{f_name}'.")

            @wraps(f)
            def decorated(*args, **kwargs):
                assert isinstance(mapping, dict), "Messed up shortcut wiring, abort."  # nosec
                logger.debug(f"Running shortcut for '{f_name}'.")

                for condition, (response, status) in mapping.items():
                    request_data = get_request_data(request)
                    try:
                        sub_resolves = diff(request_data, json.loads(condition))
                    except TypeError as e:
                        logger.debug(
                            f"Couldn't walk '{condition}' in the target request, got error message '{str(e)}'. "
                            f"This could mean that the shortcut for this function is not well-defined."
                        )
                        continue
                    if not sub_resolves:
                        continue
                    return self.app.make_response(response), status
                else:
                    logger.debug(f"Shortcut conditions couldn't be satisfied, defaulting to actual implementation.")
                return f(*args, **kwargs)

            return decorated

        # assigning the right decorator given the mapping types
        if isinstance(mapping, tuple):
            return simple_map
        elif isinstance(mapping, dict):
            return dict_map
        else:
            raise TypeError(f"'{type(mapping)}' is not a supported mapping type for shortcuts yet.")

    def wire(self, shortcuts=Dict[str, Union[RESPONSE_ARGS, Dict[str, RESPONSE_ARGS]]]):
        """Manual wiring function.

        If you don't want to have the shortcut definitions in your routing
        file for some reason (e.g. there are lots of shortcuts and it would
        make the whole thing hard to read), you can use this function at
        some point after all routes were registered, and before the server
        is started.

        Args:
            shortcuts: A dictionary that maps routes to the mappings that
                would have been used as arguments in the shortcut decorator.
        """
        route_map = {str(r): r.endpoint for r in self.app.url_map.iter_rules()}

        for route, mapping in shortcuts.items():
            if route in route_map:
                wrap = self.cut(mapping)(self.app.view_functions[route_map[route]])
                self.app.view_functions[route_map[route]] = wrap
            else:
                logger.warning(f"Can't resolve route '{route}' in the registered routes '{[*route_map]}'")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['flask_shortcut', 'tests']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.1,<8.0.0',
 'flask>=1.1.2,<2.0.0',
 'importlib-metadata>=1.6.0,<2.0.0',
 'xmltodict>=0.12.0,<0.13.0']

setup_kwargs = {
    'name': 'flask-shortcut',
    'version': '0.3.2',
    'description': 'Extension that provides an easy way to add dev-only shortcuts to your routes.',
    'long_description': '|Logo|\n\n|CI_CD| |pyPI| |Docs| |License| |py_versions| |Style|\n\n\n.. header-end\n\nProject Description\n-------------------\n\nThis extension provides an easy and safe way to add dev-only shortcuts to\nroutes in your flask application.\n\nThe main beneficiaries are microservices that need to be tested regularly in\nconjunction with their clients. If you need to assert working communication and\nbasic integration in a sufficiently complex ecosystem, clients that can not\nfreely chose how their requests are formed gain a lot from being able to\nreceive predictable responses. By skipping over the details of how the\nmicroservice is implemented, which bugs and minor changes it experiences over\ntime, testing basic API compatibility gets a lot more manageable.\n\n\n\nUsage\n-----\n\nYou can add shortcuts to your route functions either individually with\ndecorators, or in a single swoop once all routes have been defined. Both ways\nare functionally equivalent.\n\nApplying Shortcuts\n==================\n\n**With decorators:**\n\n.. code-block:: python\n\n    from flask import Flask\n    from flask_shortcut import Shortcut\n\n    app = Flask(__name__)\n    short = Shortcut(app)\n\n    app.route(\'/foo\', methods=[\'GET\'])\n    short.cut((\'short_foo\', 200))\n    def foo():\n        return \'foo\'\n\n    app.route(\'/bar\', methods=[\'POST\'])\n    short.cut({\n        \'{"name": "TestUser"}\': (\'short_bar\', 200)},\n        \'{"name": "UserTest"}\': (\'longer_bar\', 200),\n    )\n    def bar():\n        return \'bar\'\n\n**With a wire call**\n\n.. code-block:: python\n\n    from flask import Flask\n    from flask_shortcut import Shortcut\n\n    app = Flask(__name__)\n\n    app.route(\'/foo\', methods=[\'GET\'])\n    def foo():\n        return \'foo\'\n\n    app.route(\'/bar\', methods=[\'POST\'])\n    def bar():\n        return \'bar\'\n\n    Shortcut(app).wire(\n        {\n             \'/foo\': (\'short_foo\', 200),\n             \'/bar\': {\n                 \'{"name": "TestUser"}\': (\'short_bar\', 200),\n                 \'{"name": "UserTest"}\': (\'longer_bar\', 200),\n             }\n        }\n    )\n\n\nWhat it looks like\n==================\n\nTo showcase how the shortcuts are supposed to work, here is the result\nof a couple of requests sent against the server from the example above\nif it were run with ``FLASK_ENV=test flask run``:\n\n.. code-block:: python\n\n  >>> from request import get, post\n  >>> get(\'http://127.0.0.1:5000/foo\').text\n  \'short_foo\'  # the only response this route will give\n  >>> post(\'http://127.0.0.1:5000/bar\', json={"name": "me"}).text\n  \'bar\'  # no shortcut match -> the original logic was executed\n  >>> post(\'http://127.0.0.1:5000/bar\', json={"name": "TestUser"}).text\n  \'short_bar\'  # shortcut match\n  >>> post(\'http://127.0.0.1:5000/bar\', json={"name": "UserTest", "job": None}).text\n  \'longer_bar\'  # shortcut only needs to be contained for a match\n\nOne focus of this package was, that a production deployment would remain\nas ignorant as possible about the existence of shortcuts. While the\nshortcut object is still created, it only delegates the route functions\nand no shortcut code has any chance of being run.\n\n\nConfiguration\n-------------\n\nBy default, shortcuts will only be applied when ``FLASK_ENV`` is set to\nsomething different than the default setting ``production``. You can\nextend that list through the ``SHORTCUT_EXCLUSIONS`` config setting,\neither by adding it to your app\'s config before creating any Shortcut\nobjects, or preferably by setting up the whole config `through a file`_.\n\nPossible values for it are all environments other than ``production`` that\nyou want to block separated by commas, for example ``staging,master``.\n\n----\n\nProject home is `on github`_.\n\n.. |Logo| image:: https://user-images.githubusercontent.com/2063412/79631833-c1b39400-815b-11ea-90da-d9264420ef68.png\n   :alt: Logo\n   :width: 1200\n   :target: https://github.com/a-recknagel/Flask-Shortcut\n\n.. |CI_CD| image:: https://github.com/a-recknagel/Flask-Shortcut/workflows/CI-CD/badge.svg\n   :alt: Main workflow status\n   :target: https://github.com/a-recknagel/Flask-Shortcut/actions\n\n.. |pyPI| image:: https://img.shields.io/pypi/v/flask-shortcut\n   :alt: Current pyPI version\n   :target: https://pypi.org/project/flask-shortcut/\n\n.. |Docs| image:: https://img.shields.io/badge/docs-github--pages-blue\n   :alt: Documentation home\n   :target: https://a-recknagel.github.io/Flask-Shortcut/\n\n.. |License| image:: https://img.shields.io/pypi/l/flask-shortcut\n   :alt: Package license\n   :target: https://pypi.org/project/flask-shortcut/\n\n.. |py_versions| image:: https://img.shields.io/pypi/pyversions/flask-shortcut\n   :alt: Supported on python versions\n   :target: https://pypi.org/project/flask-shortcut/\n\n.. |Style| image:: https://img.shields.io/badge/codestyle-black-black\n   :alt: Any color you want\n   :target: https://black.readthedocs.io/en/stable/\n\n.. _on github: https://github.com/a-recknagel/Flask-Shortcut\n\n.. _through a file: https://flask.palletsprojects.com/en/1.1.x/config/#configuring-from-files',
    'author': 'Arne',
    'author_email': 'arecknag@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/a-recknagel/Flask-Shortcut',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from typing import Any

from flask import Request
import xmltodict


def get_request_data(r: Request) -> Any:
    """Request data fetcher.

    This function inspects the mimetype in order to figure out how the data
    can be represented in a standard python-dictionary form.

    Args:
        r: The request whose body is going to be parsed.

    Returns:
        The data in a form that sticks as close a parsed json object as possible, since
            that is the format in which the mappings expect to be passed by.

    Raises:
        ValueError: In case of unexpected request data formats.
        Exception: In case of junk data, hard to list what each parser
            decides is best to raise.
    """
    if "json" in r.mimetype:
        return r.json
    if "xml" in r.mimetype:
        return xmltodict.parse(r.data, dict_constructor=dict)
    raise ValueError(f"Mimetype '{r.mimetype}' not parsable.")


def diff(target, sub, path_=None) -> bool:
    """Simple recursive asymmetric diff function on json-like data structures.

    Args:
        target: The data container in which we want to find a subtree.
        sub: The subtree that needs to be matched.
        path_: A private object used to produce meaningful error messages.

    Returns:
        True if the subtree could be cleanly matched, else False.

    Raises:
        TypeError: If the parse paths don't match, e.g. the target is a list
            at a level where the subtree expected it to be a dictionary.
    """
    if path_ is None:
        path_ = ["root"]

    if type(target) != type(sub):
        raise TypeError(f"Types of target '{type(target)}' and sub '{type(sub)}' are incompatible at junction {path_}.")

    if isinstance(sub, list):
        for n, s_elem in enumerate(sub):
            for t_elem in target:
                if diff(t_elem, s_elem, [*path_, n]):
                    break
            else:
                return False
        return True
    elif isinstance(sub, dict):
        for s_key in sub:
            if s_key in target and diff(target[s_key], sub[s_key], [*path_, s_key]):
                continue
            return False
        return True
    else:  # some kind of leaf, i.e. string, int, bool, float, or None
        return bool(target == sub)
